@@ -1,10 +1,12 @@
+from functools import wraps
+import os
 import sys
 
 # make sure gevent-socketio is not installed, as it conflicts with
 # python-socketio
 gevent_socketio_found = True
 try:
-    from socketio import socketio_manage
+    from socketio import socketio_manage  # noqa: F401
 except ImportError:
     gevent_socketio_found = False
 if gevent_socketio_found:
@@ -13,19 +15,21 @@ if gevent_socketio_found:
           'install the latest version of python-socketio in its place.')
     sys.exit(1)
 
-import socketio
 import flask
-from flask import _request_ctx_stack, json as flask_json
+from flask import _request_ctx_stack, has_request_context, json as flask_json
+from flask.sessions import SessionMixin
+import socketio
+from socketio.exceptions import ConnectionRefusedError  # noqa: F401
 from werkzeug.debug import DebuggedApplication
 from werkzeug.serving import run_with_reloader
 
 from .namespace import Namespace
 from .test_client import SocketIOTestClient
 
-__version__ = '2.9.0'
+__version__ = '5.0.1'
 
 
-class _SocketIOMiddleware(socketio.Middleware):
+class _SocketIOMiddleware(socketio.WSGIApp):
     """This WSGI middleware simply exposes the Flask application in the WSGI
     environment before executing the request.
     """
@@ -33,7 +37,7 @@ class _SocketIOMiddleware(socketio.Middleware):
         self.flask_app = flask_app
         super(_SocketIOMiddleware, self).__init__(socketio_app,
                                                   flask_app.wsgi_app,
-                                                  socketio_path)
+                                                  socketio_path=socketio_path)
 
     def __call__(self, environ, start_response):
         environ = environ.copy()
@@ -42,6 +46,13 @@ class _SocketIOMiddleware(socketio.Middleware):
                                                          start_response)
 
 
+class _ManagedSession(dict, SessionMixin):
+    """This class is used for user sessions that are managed by
+    Flask-SocketIO. It is simple dict, expanded with the Flask session
+    attributes."""
+    pass
+
+
 class SocketIO(object):
     """Create a Flask-SocketIO server.
 
@@ -64,8 +75,8 @@ class SocketIO(object):
     :param channel: The channel name, when using a message queue. If a channel
                     isn't specified, a default channel will be used. If
                     multiple clusters of SocketIO processes need to use the
-                    same message queue without interfering with each other, then
-                    each cluster should use a different channel.
+                    same message queue without interfering with each other,
+                    then each cluster should use a different channel.
     :param path: The path where the Socket.IO server is exposed. Defaults to
                  ``'socket.io'``. Leave this as is unless you know what you are
                  doing.
@@ -81,7 +92,9 @@ class SocketIO(object):
                            cases, this argument does not need to be set
                            explicitly.
     :param logger: To enable logging set to ``True`` or pass a logger object to
-                   use. To disable logging set to ``False``.
+                   use. To disable logging set to ``False``. The default is
+                   ``False``. Note that fatal errors will be logged even when
+                   ``logger`` is ``False``.
     :param binary: ``True`` to support binary payloads, ``False`` to treat all
                    payloads as text. On Python 2, if this is set to ``True``,
                    ``unicode`` values are treated as text, and ``str`` and
@@ -93,6 +106,20 @@ class SocketIO(object):
                  functions that are compatible with the standard library
                  versions. To use the same json encoder and decoder as a Flask
                  application, use ``flask.json``.
+    :param async_handlers: If set to ``True``, event handlers for a client are
+                           executed in separate threads. To run handlers for a
+                           client synchronously, set to ``False``. The default
+                           is ``True``.
+    :param always_connect: When set to ``False``, new connections are
+                           provisory until the connect handler returns
+                           something other than ``False``, at which point they
+                           are accepted. When set to ``True``, connections are
+                           immediately accepted, and then if the connect
+                           handler returns ``False`` a disconnect is issued.
+                           Set to ``True`` if you need to emit events from the
+                           connect handler and your client is confused when it
+                           receives events before the connection acceptance.
+                           In any other case use the default of ``False``.
 
     The Engine.IO server configuration supports the following settings:
 
@@ -106,26 +133,40 @@ class SocketIO(object):
                        first async mode that has all its dependencies installed
                        is then one that is chosen.
     :param ping_timeout: The time in seconds that the client waits for the
-                         server to respond before disconnecting.
+                         server to respond before disconnecting. The default is
+                         60 seconds.
     :param ping_interval: The interval in seconds at which the client pings
-                          the server.
+                          the server. The default is 25 seconds.
     :param max_http_buffer_size: The maximum size of a message when using the
-                                 polling transport.
-    :param allow_upgrades: Whether to allow transport upgrades or not.
+                                 polling transport. The default is 100,000,000
+                                 bytes.
+    :param allow_upgrades: Whether to allow transport upgrades or not. The
+                           default is ``True``.
     :param http_compression: Whether to compress packages when using the
-                             polling transport.
+                             polling transport. The default is ``True``.
     :param compression_threshold: Only compress messages when their byte size
-                                  is greater than this value.
+                                  is greater than this value. The default is
+                                  1024 bytes.
     :param cookie: Name of the HTTP cookie that contains the client session
                    id. If set to ``None``, a cookie is not sent to the client.
-    :param cors_allowed_origins: List of origins that are allowed to connect
-                                 to this server. All origins are allowed by
-                                 default.
+                   The default is ``'io'``.
+    :param cors_allowed_origins: Origin or list of origins that are allowed to
+                                 connect to this server. Only the same origin
+                                 is allowed by default. Set this argument to
+                                 ``'*'`` to allow all origins, or to ``[]`` to
+                                 disable CORS handling.
     :param cors_credentials: Whether credentials (cookies, authentication) are
-                             allowed in requests to this server.
+                             allowed in requests to this server. The default is
+                             ``True``.
+    :param monitor_clients: If set to ``True``, a background task will ensure
+                            inactive clients are closed. Set to ``False`` to
+                            disable the monitoring task (not recommended). The
+                            default is ``True``.
     :param engineio_logger: To enable Engine.IO logging set to ``True`` or pass
                             a logger object to use. To disable logging set to
-                            ``False``.
+                            ``False``. The default is ``False``. Note that
+                            fatal errors are logged even when
+                            ``engineio_logger`` is ``False``.
     """
 
     def __init__(self, app=None, **kwargs):
@@ -162,8 +203,10 @@ class SocketIO(object):
             channel = self.server_options.pop('channel', 'flask-socketio')
             write_only = app is None
             if url:
-                if url.startswith('redis://'):
+                if url.startswith(('redis://', "rediss://")):
                     queue_class = socketio.RedisManager
+                elif url.startswith(('kafka://')):
+                    queue_class = socketio.KafkaManager
                 elif url.startswith('zmq'):
                     queue_class = socketio.ZmqManager
                 else:
@@ -195,6 +238,9 @@ class SocketIO(object):
             self.server_options.pop('resource', None) or 'socket.io'
         if resource.startswith('/'):
             resource = resource[1:]
+        if os.environ.get('FLASK_RUN_FROM_CLI'):
+            if self.server_options.get('async_mode') is None:
+                self.server_options['async_mode'] = 'threading'
         self.server = socketio.Server(**self.server_options)
         self.async_mode = self.server.async_mode
         for handler in self.handlers:
@@ -203,8 +249,9 @@ class SocketIO(object):
             self.server.register_namespace(namespace_handler)
 
         if app is not None:
-            # here we attach the SocketIO middlware to the SocketIO object so it
-            # can be referenced later if debug middleware needs to be inserted
+            # here we attach the SocketIO middlware to the SocketIO object so
+            # it can be referenced later if debug middleware needs to be
+            # inserted
             self.sockio_mw = _SocketIOMiddleware(self.server, app,
                                                  socketio_path=resource)
             app.wsgi_app = self.sockio_mw
@@ -231,6 +278,7 @@ class SocketIO(object):
         namespace = namespace or '/'
 
         def decorator(handler):
+            @wraps(handler)
             def _handler(sid, *args):
                 return self._handle_event(handler, message, namespace, sid,
                                           *args)
@@ -307,6 +355,41 @@ class SocketIO(object):
         """
         self.on(message, namespace=namespace)(handler)
 
+    def event(self, *args, **kwargs):
+        """Decorator to register an event handler.
+
+        This is a simplified version of the ``on()`` method that takes the
+        event name from the decorated function.
+
+        Example usage::
+
+            @socketio.event
+            def my_event(data):
+                print('Received data: ', data)
+
+        The above example is equivalent to::
+
+            @socketio.on('my_event')
+            def my_event(data):
+                print('Received data: ', data)
+
+        A custom namespace can be given as an argument to the decorator::
+
+            @socketio.event(namespace='/test')
+            def my_event(data):
+                print('Received data: ', data)
+        """
+        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
+            # the decorator was invoked without arguments
+            # args[0] is the decorated function
+            return self.on(args[0].__name__)(args[0])
+        else:
+            # the decorator was invoked with arguments
+            def set_handler(handler):
+                return self.on(handler.__name__, *args, **kwargs)(handler)
+
+            return set_handler
+
     def on_namespace(self, namespace_handler):
         if not isinstance(namespace_handler, Namespace):
             raise ValueError('Not a namespace instance.')
@@ -334,13 +417,17 @@ class SocketIO(object):
         :param args: A dictionary with the JSON data to send as payload.
         :param namespace: The namespace under which the message is to be sent.
                           Defaults to the global namespace.
-        :param room: Send the message to all the users in the given room. If
-                     this parameter is not included, the event is sent to
-                     all connected users.
+        :param to: Send the message to all the users in the given room. If
+                   this parameter is not included, the event is sent to all
+                   connected users.
+        :param include_self: ``True`` to include the sender when broadcasting
+                             or addressing a room, or ``False`` to send to
+                             everyone but the sender.
         :param skip_sid: The session id of a client to ignore when broadcasting
                          or addressing a room. This is typically set to the
                          originator of the message, so that everyone except
-                         that client receive the message.
+                         that client receive the message. To skip multiple sids
+                         pass a list.
         :param callback: If given, this function will be called to acknowledge
                          that the client has received the message. The
                          arguments that will be passed to the function are
@@ -348,16 +435,33 @@ class SocketIO(object):
                          only be used when addressing an individual client.
         """
         namespace = kwargs.pop('namespace', '/')
-        room = kwargs.pop('room', None)
+        to = kwargs.pop('to', kwargs.pop('room', None))
         include_self = kwargs.pop('include_self', True)
         skip_sid = kwargs.pop('skip_sid', None)
         if not include_self and not skip_sid:
             skip_sid = flask.request.sid
         callback = kwargs.pop('callback', None)
-        self.server.emit(event, *args, namespace=namespace, room=room,
+        if callback:
+            # wrap the callback so that it sets app app and request contexts
+            sid = None
+            if has_request_context():
+                sid = getattr(flask.request, 'sid', None)
+            original_callback = callback
+
+            def _callback_wrapper(*args):
+                return self._handle_event(original_callback, None, namespace,
+                                          sid, *args)
+
+            if sid:
+                # the callback wrapper above will install a request context
+                # before invoking the original callback
+                # we only use it if the emit was issued from a Socket.IO
+                # populated request context (i.e. request.sid is defined)
+                callback = _callback_wrapper
+        self.server.emit(event, *args, namespace=namespace, to=to,
                          skip_sid=skip_sid, callback=callback, **kwargs)
 
-    def send(self, data, json=False, namespace=None, room=None,
+    def send(self, data, json=False, namespace=None, to=None,
              callback=None, include_self=True, skip_sid=None, **kwargs):
         """Send a server-generated SocketIO message.
 
@@ -367,18 +471,22 @@ class SocketIO(object):
         used outside of a SocketIO event context, so it is appropriate to use
         when the server is the originator of an event.
 
-        :param message: The message to send, either a string or a JSON blob.
+        :param data: The message to send, either a string or a JSON blob.
         :param json: ``True`` if ``message`` is a JSON blob, ``False``
                      otherwise.
         :param namespace: The namespace under which the message is to be sent.
                           Defaults to the global namespace.
-        :param room: Send the message only to the users in the given room. If
-                     this parameter is not included, the message is sent to
-                     all connected users.
+        :param to: Send the message only to the users in the given room. If
+                   this parameter is not included, the message is sent to all
+                   connected users.
+        :param include_self: ``True`` to include the sender when broadcasting
+                             or addressing a room, or ``False`` to send to
+                             everyone but the sender.
         :param skip_sid: The session id of a client to ignore when broadcasting
                          or addressing a room. This is typically set to the
                          originator of the message, so that everyone except
-                         that client receive the message.
+                         that client receive the message. To skip multiple sids
+                         pass a list.
         :param callback: If given, this function will be called to acknowledge
                          that the client has received the message. The
                          arguments that will be passed to the function are
@@ -387,10 +495,10 @@ class SocketIO(object):
         """
         skip_sid = flask.request.sid if not include_self else skip_sid
         if json:
-            self.emit('json', data, namespace=namespace, room=room,
+            self.emit('json', data, namespace=namespace, to=to,
                       skip_sid=skip_sid, callback=callback, **kwargs)
         else:
-            self.emit('message', data, namespace=namespace, room=room,
+            self.emit('message', data, namespace=namespace, to=to,
                       skip_sid=skip_sid, callback=callback, **kwargs)
 
     def close_room(self, room, namespace=None):
@@ -469,8 +577,8 @@ class SocketIO(object):
             #     o
             #  Flask-SocketIO WebSocket handler
             #
-            self.sockio_mw.wsgi_app = DebuggedApplication(self.sockio_mw.wsgi_app,
-                                                          evalex=True)
+            self.sockio_mw.wsgi_app = DebuggedApplication(
+                self.sockio_mw.wsgi_app, evalex=True)
 
         if self.server.eio.async_mode == 'threading':
             from werkzeug._internal import _log
@@ -486,8 +594,10 @@ class SocketIO(object):
                 import eventlet.green
                 addresses = eventlet.green.socket.getaddrinfo(host, port)
                 if not addresses:
-                    raise RuntimeError('Could not resolve host to a valid address')
-                eventlet_socket = eventlet.listen(addresses[0][4], addresses[0][0])
+                    raise RuntimeError(
+                        'Could not resolve host to a valid address')
+                eventlet_socket = eventlet.listen(addresses[0][4],
+                                                  addresses[0][0])
 
                 # If provided an SSL argument, use an SSL socket
                 ssl_args = ['keyfile', 'certfile', 'server_side', 'cert_reqs',
@@ -515,6 +625,9 @@ class SocketIO(object):
                 from geventwebsocket.handler import WebSocketHandler
                 websocket = True
             except ImportError:
+                app.logger.warning(
+                    'WebSocket transport not available. Install '
+                    'gevent-websocket for improved performance.')
                 websocket = False
 
             log = 'default'
@@ -526,12 +639,13 @@ class SocketIO(object):
                     log=log, **kwargs)
             else:
                 self.wsgi_server = pywsgi.WSGIServer((host, port), app,
-                                                     log=log)
+                                                     log=log, **kwargs)
 
             if use_reloader:
                 # monkey patching is required by the reloader
                 from gevent import monkey
-                monkey.patch_all()
+                monkey.patch_thread()
+                monkey.patch_time()
 
                 def run_server():
                     self.wsgi_server.serve_forever()
@@ -583,23 +697,42 @@ class SocketIO(object):
         """
         return self.server.sleep(seconds)
 
-    def test_client(self, app, namespace=None):
-        """Return a simple SocketIO client that can be used for unit tests."""
-        return SocketIOTestClient(app, self, namespace)
+    def test_client(self, app, namespace=None, query_string=None,
+                    headers=None, flask_test_client=None):
+        """The Socket.IO test client is useful for testing a Flask-SocketIO
+        server. It works in a similar way to the Flask Test Client, but
+        adapted to the Socket.IO server.
+
+        :param app: The Flask application instance.
+        :param namespace: The namespace for the client. If not provided, the
+                          client connects to the server on the global
+                          namespace.
+        :param query_string: A string with custom query string arguments.
+        :param headers: A dictionary with custom HTTP headers.
+        :param flask_test_client: The instance of the Flask test client
+                                  currently in use. Passing the Flask test
+                                  client is optional, but is necessary if you
+                                  want the Flask user session and any other
+                                  cookies set in HTTP routes accessible from
+                                  Socket.IO events.
+        """
+        return SocketIOTestClient(app, self, namespace=namespace,
+                                  query_string=query_string, headers=headers,
+                                  flask_test_client=flask_test_client)
 
     def _handle_event(self, handler, message, namespace, sid, *args):
-        if sid not in self.server.environ:
+        environ = self.server.get_environ(sid, namespace=namespace)
+        if not environ:
             # we don't have record of this client, ignore this event
             return '', 400
-        app = self.server.environ[sid]['flask.app']
-        with app.request_context(self.server.environ[sid]):
+        app = environ['flask.app']
+        with app.request_context(environ):
             if self.manage_session:
                 # manage a separate session for this client's Socket.IO events
                 # created as a copy of the regular user session
-                if 'saved_session' not in self.server.environ[sid]:
-                    self.server.environ[sid]['saved_session'] = \
-                        dict(flask.session)
-                session_obj = self.server.environ[sid]['saved_session']
+                if 'saved_session' not in environ:
+                    environ['saved_session'] = _ManagedSession(flask.session)
+                session_obj = environ['saved_session']
             else:
                 # let Flask handle the user session
                 # for cookie based sessions, this effectively freezes the
@@ -625,7 +758,8 @@ class SocketIO(object):
                 return err_handler(value)
             if not self.manage_session:
                 # when Flask is managing the user session, it needs to save it
-                if not hasattr(session_obj, 'modified') or session_obj.modified:
+                if not hasattr(session_obj, 'modified') or \
+                        session_obj.modified:
                     resp = app.response_class()
                     app.session_interface.save_session(app, session_obj, resp)
             return ret
@@ -653,11 +787,24 @@ def emit(event, *args, **kwargs):
                      acknowledgement.
     :param broadcast: ``True`` to send the message to all clients, or ``False``
                       to only reply to the sender of the originating event.
-    :param room: Send the message to all the users in the given room. If this
-                 argument is set, then broadcast is implied to be ``True``.
+    :param to: Send the message to all the users in the given room. If this
+               argument is not set and ``broadcast`` is ``False``, then the
+               message is sent only to the originating user.
     :param include_self: ``True`` to include the sender when broadcasting or
                          addressing a room, or ``False`` to send to everyone
                          but the sender.
+    :param skip_sid: The session id of a client to ignore when broadcasting
+                     or addressing a room. This is typically set to the
+                     originator of the message, so that everyone except
+                     that client receive the message. To skip multiple sids
+                     pass a list.
+    :param ignore_queue: Only used when a message queue is configured. If
+                         set to ``True``, the event is emitted to the
+                         clients directly, without going through the queue.
+                         This is more efficient, but only works when a
+                         single server process is used, or when there is a
+                         single addressee. It is recommended to always leave
+                         this parameter with its default value of ``False``.
     """
     if 'namespace' in kwargs:
         namespace = kwargs['namespace']
@@ -665,14 +812,17 @@ def emit(event, *args, **kwargs):
         namespace = flask.request.namespace
     callback = kwargs.get('callback')
     broadcast = kwargs.get('broadcast')
-    room = kwargs.get('room')
-    if room is None and not broadcast:
-        room = flask.request.sid
+    to = kwargs.pop('to', kwargs.pop('room', None))
+    if to is None and not broadcast:
+        to = flask.request.sid
     include_self = kwargs.get('include_self', True)
+    skip_sid = kwargs.get('skip_sid')
+    ignore_queue = kwargs.get('ignore_queue', False)
 
     socketio = flask.current_app.extensions['socketio']
-    return socketio.emit(event, *args, namespace=namespace, room=room,
-                         include_self=include_self, callback=callback)
+    return socketio.emit(event, *args, namespace=namespace, to=to,
+                         include_self=include_self, skip_sid=skip_sid,
+                         callback=callback, ignore_queue=ignore_queue)
 
 
 def send(message, **kwargs):
@@ -684,6 +834,8 @@ def send(message, **kwargs):
     can only be called from a SocketIO event handler.
 
     :param message: The message to send, either a string or a JSON blob.
+    :param json: ``True`` if ``message`` is a JSON blob, ``False``
+                     otherwise.
     :param namespace: The namespace under which the message is to be sent.
                       Defaults to the namespace used by the originating event.
                       An empty string can be used to use the global namespace.
@@ -692,25 +844,43 @@ def send(message, **kwargs):
     :param broadcast: ``True`` to send the message to all connected clients, or
                       ``False`` to only reply to the sender of the originating
                       event.
-    :param room: Send the message to all the users in the given room.
+    :param to: Send the message to all the users in the given room. If this
+               argument is not set and ``broadcast`` is ``False``, then the
+               message is sent only to the originating user.
     :param include_self: ``True`` to include the sender when broadcasting or
                          addressing a room, or ``False`` to send to everyone
                          but the sender.
+    :param skip_sid: The session id of a client to ignore when broadcasting
+                     or addressing a room. This is typically set to the
+                     originator of the message, so that everyone except
+                     that client receive the message. To skip multiple sids
+                     pass a list.
+    :param ignore_queue: Only used when a message queue is configured. If
+                         set to ``True``, the event is emitted to the
+                         clients directly, without going through the queue.
+                         This is more efficient, but only works when a
+                         single server process is used, or when there is a
+                         single addressee. It is recommended to always leave
+                         this parameter with its default value of ``False``.
     """
+    json = kwargs.get('json', False)
     if 'namespace' in kwargs:
         namespace = kwargs['namespace']
     else:
         namespace = flask.request.namespace
     callback = kwargs.get('callback')
     broadcast = kwargs.get('broadcast')
-    room = kwargs.get('room')
-    if room is None and not broadcast:
-        room = flask.request.sid
+    to = kwargs.pop('to', kwargs.pop('room', None))
+    if to is None and not broadcast:
+        to = flask.request.sid
     include_self = kwargs.get('include_self', True)
+    skip_sid = kwargs.get('skip_sid')
+    ignore_queue = kwargs.get('ignore_queue', False)
 
     socketio = flask.current_app.extensions['socketio']
-    return socketio.send(message, namespace=namespace, room=room,
-                         include_self=include_self, callback=callback)
+    return socketio.send(message, json=json, namespace=namespace, to=to,
+                         include_self=include_self, skip_sid=skip_sid,
+                         callback=callback, ignore_queue=ignore_queue)
 
 
 def join_room(room, sid=None, namespace=None):
@@ -796,7 +966,7 @@ def rooms(sid=None, namespace=None):
     return socketio.server.rooms(sid, namespace=namespace)
 
 
-def disconnect(silent=False):
+def disconnect(sid=None, namespace=None, silent=False):
     """Disconnect the client.
 
     This function terminates the connection with the client. As a result of
@@ -806,10 +976,16 @@ def disconnect(silent=False):
         def receive_message(msg):
             if is_banned(session['username']):
                 disconnect()
-            # ...
+            else:
+                # ...
 
+    :param sid: The session id of the client. If not provided, the client is
+                obtained from the request context.
+    :param namespace: The namespace for the room. If not provided, the
+                      namespace is obtained from the request context.
     :param silent: this option is deprecated.
     """
     socketio = flask.current_app.extensions['socketio']
-    return socketio.server.disconnect(flask.request.sid,
-                                      namespace=flask.request.namespace)
+    sid = sid or flask.request.sid
+    namespace = namespace or flask.request.namespace
+    return socketio.server.disconnect(sid, namespace=namespace)
